from typing import Union,Literal,List,Dict
from datetime import datetime
from skynamo.models.Address import Address
from skynamo.shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from skynamo.models.Location import Location
from skynamo.reader.readerHelpers import getCustomFieldWithName

class Customer:
	def getCustomFieldWithName(self,name:str):
		return getCustomFieldWithName(self,name)
	def __init__(self,json:dict):
		self.id:int=json['id']
		self.code:str=json['code']
		self.name:str=json['name']
		self.active:bool=json['active']
		self.location:Location=Location(**json['location'])
		self.price_list_id:Union[int,None]=None
		if 'price_list_id' in json:
			self.price_list_id=json['price_list_id']

		self.price_list_name:Union[str,None]=None
		if 'price_list_name' in json:
			self.price_list_name=json['price_list_name']
		self.assigned_users:List[int]=json['assigned_users']
		self.default_discount:float=json['default_discount']
		self.default_warehouse_id:Union[int,None]=None
		if 'default_warehouse_id' in json:
			self.default_warehouse_id=json['default_warehouse_id']
		self.default_warehouse_name:Union[str,None]=None
		if 'default_warehouse_name' in json:
			self.default_warehouse_name=json['default_warehouse_name']
		self.row_version:int=json['version']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])
		self.create_date:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['create_date'])





from typing import Literal

class CustomFieldsToCreate:
	def __init__(self):
		self.fields_to_add=[]
##fieldtypes=Text, Number, SingleSelect, MultiSelect, NestedSingleSelect, NestedMultiSelect, Address, UserSingleSelect, UserMultiSelect
	def addCustomerCustomField(self,name:str,type:Literal['Text','Number','SingleSelect','MultiSelect','NestedSingleSelect','NestedMultiSelect','Address']):
		self.fields_to_add.append({'name':name,'type':type,'form_id':-1})
	def addProductCustomField(self,name:str,type:Literal['Text','Number','SingleSelect','MultiSelect','NestedSingleSelect','NestedMultiSelect','Address']):
		self.fields_to_add.append({'name':name,'type':type,'form_id':-3})
	def addFormCustomField(self,formId:int,name:str,type:Literal['Text','Number','SingleSelect','MultiSelect','NestedSingleSelect','NestedMultiSelect','Address']):
		self.fields_to_add.append({'name':name,'type':type,'form_id':formId})


from typing import Literal
class VisitFrequency:
    def __init__(self,json:dict):
        self.id: int = json['id']
        self.customer_id: int = json['customer_id']
        self.customer_code: str = json['customer_code']
        self.customer_name: str = json['customer_name']
        self.user_id: int = json['user_id']
        self.user_name: str = json['user_name']
        self.numberOfCyclesPerPeriod:int = json['cycle']
        self.numberOfVisitsRequiredPerCycle:int = json['frequency']
        self.period:Literal['Week','Month','Year']= json['period']
        self.version:int = json['version']
        self.last_modified_time:str = json['last_modified_time']



from .models.Address import Address
from .models.OrderUnit import OrderUnit
from .models.InvoiceItem import InvoiceItem
from .models.Location import Location
from .models.LineItem import LineItem
from .models.CustomFieldsToCreate import CustomFieldsToCreate
from .refresher import refreshCustomFormsAndFields
from .outputters.csvWriter import writeListOfObjectsToCsvWithObjectPropertiesAsColumnNames
from .outputters.emailer import sendEmailUsingGmailCredentialsWithFilesAttached
from .main.Writer import Writer
from .main.Reader import Reader
try:
    from skynamo_data.code.Reader import Reader
    from skynamo_data.code.Writer import Writer
except:
    pass


from ..shared.helpers import writeToInstanceSpecificFile,clearInstanceSpecificFolder
import os
from .refreshHelpers import getNameOfModelClassOnWhichToAddCustomFieldFromFormDefinition,isCustomForm
from .CustomFieldArg import getListCustomFieldArgs,CustomFieldArg
from typing import List,Dict

def getStringContentOfClassInMainFolder(className:str):
	if className[-3:]!='.py':
		className+='.py'
	sharedPath = os.path.dirname(os.path.realpath(__file__)).replace('\\','/')
	mainPath=sharedPath.replace('/refresher','/main')
	with open(f'{mainPath}/{className}', "r") as read_file:
		return read_file.read()

class InstanceClassContents:
	def __init__(self):
		self.writerContent=getStringContentOfClassInMainFolder('Writer')
		self.readerContent=getStringContentOfClassInMainFolder('Reader')
		self.orderContent=getStringContentOfClassInMainFolder('Order')
		self.quoteContent=getStringContentOfClassInMainFolder('Quote')
		self.creditrequestContent=getStringContentOfClassInMainFolder('CreditRequest')
		self.productContent=getStringContentOfClassInMainFolder('Product')
		self.customerContent=getStringContentOfClassInMainFolder('Customer')

		self.__customFormBaseContent=getStringContentOfClassInMainFolder('CustomForm')
		self.customFormsContent={}
		self.modelNameToCustomFieldPropTypes={}

	def write(self):
		self.addCustomFieldPropTypesToRelevantClassContents()
		clearInstanceSpecificFolder('code')
		writeToInstanceSpecificFile('code/Writer.py',self.writerContent)
		writeToInstanceSpecificFile('code/Reader.py',self.readerContent)
		writeToInstanceSpecificFile('code/Order.py',self.orderContent)
		writeToInstanceSpecificFile('code/Quote.py',self.quoteContent)
		writeToInstanceSpecificFile('code/CreditRequest.py',self.creditrequestContent)
		writeToInstanceSpecificFile('code/Product.py',self.productContent)
		writeToInstanceSpecificFile('code/Customer.py',self.customerContent)
		for customFormName,customFormContent in self.customFormsContent.items():
			writeToInstanceSpecificFile(f'code/{customFormName}.py',customFormContent)

	def addCustomFieldPropTypesToRelevantClassContents(self):
		standardClasses=['Order','Quote','CreditRequest','Product','Customer']
		for i,modelClassName in enumerate(standardClasses):
			classContent=self.__getattribute__(f'{modelClassName.lower()}Content')
			customFieldArgsWithPropTypes={}
			if modelClassName in self.modelNameToCustomFieldPropTypes:
				customFieldArgsWithPropTypes=self.modelNameToCustomFieldPropTypes[modelClassName]
			classContent+=f'\t\tself._customFieldPropTypes:Dict[str,str]={customFieldArgsWithPropTypes}\r'
			self.__updateClassContents(modelClassName,classContent)
		for customFormName,customFormContent in self.customFormsContent.items():
			customFieldArgsWithPropTypes=self.modelNameToCustomFieldPropTypes[customFormName]
			customFormContent+=f'\t\tself._customFieldPropTypes:Dict[str,str]={customFieldArgsWithPropTypes}\r'
			self.customFormsContent[customFormName]=customFormContent

	def addFormDefinitionToRelevantClassContents(self,formDef:dict):
		formId=formDef['id']
		formName=formDef['name']
		if formName in ['Dropbox','SkynamoServices']:
			return
		customFieldArgs=getListCustomFieldArgs(formDef)
		modelClassName=getNameOfModelClassOnWhichToAddCustomFieldFromFormDefinition(formDef)
		self.__addCustomFieldsToRelevantModelClassContents(customFieldArgs,modelClassName,formId)
		self.__addCustomFieldsToWriterIfApplicable(customFieldArgs,modelClassName)
		self.__addCustomFormGetMethodsToReader(modelClassName)

	def __addCustomFieldsToRelevantModelClassContents(self,customFieldArgs:List[CustomFieldArg],modelClassName:str,formId:str):
		modelClassString=self.__customFormBaseContent.replace(f'class CustomForm(',f'class {modelClassName}(')
		if not(isCustomForm(modelClassName)):
			modelClassString=self.__getattribute__(f'{modelClassName.lower()}Content')
		customFieldArgsWithPropTypes={}
		for customFieldArg in customFieldArgs:
			modelClassString+=f'\t\tself.{customFieldArg.argName}:Union[{customFieldArg.argType},None]=None\r'
			customFieldArgsWithPropTypes[customFieldArg.argName]=customFieldArg.argType
		if modelClassName not in self.modelNameToCustomFieldPropTypes:
			self.modelNameToCustomFieldPropTypes[modelClassName]={}
		for key in customFieldArgsWithPropTypes:
			self.modelNameToCustomFieldPropTypes[modelClassName][key]=customFieldArgsWithPropTypes[key]
		self.__updateClassContents(modelClassName,modelClassString)

	def __updateClassContents(self,modelClassName:str,modelClassString:str):
		if not(isCustomForm(modelClassName)):
			self.__setattr__(f'{modelClassName.lower()}Content',modelClassString)
		else:
			self.customFormsContent[modelClassName]=modelClassString

	def __addCustomFieldsToWriterIfApplicable(self,customFieldArgs:List[CustomFieldArg],modelClassName:str):
		if not(isCustomForm(modelClassName)):
			requiredCustomFieldsArgsAsString=[]
			optionalCustomFieldsArgsAsString=[]
			for customFieldArg in customFieldArgs:
				if customFieldArg.required:
					requiredCustomFieldsArgsAsString.append(f'{customFieldArg.argName}:{customFieldArg.argType}')
				else:
					propType=f'Union[{customFieldArg.argType},None]'
					if customFieldArg.argType=='None':
						propType='None'
					optionalCustomFieldsArgsAsString.append(f'{customFieldArg.argName}:{propType}=None')
			placeHolder=f'##|required{modelClassName}CustomFields|##'
			replacement=placeHolder
			if len(requiredCustomFieldsArgsAsString)>0:
				replacement=','+replacement
			self.writerContent=self.writerContent.replace(placeHolder,','.join(requiredCustomFieldsArgsAsString)+replacement)
			placeHolder=f'##|optional{modelClassName}CustomFields|##'
			replacement=placeHolder
			if len(optionalCustomFieldsArgsAsString)>0:
				replacement=','+replacement
			self.writerContent=self.writerContent.replace(placeHolder,','.join(optionalCustomFieldsArgsAsString)+replacement)

	def __addCustomFormGetMethodsToReader(self,modelClassName:str):
		if isCustomForm(modelClassName):
			importString=f'from .{modelClassName} import {modelClassName}\r'
			importPlaceHolder='##|customImports|##'
			self.readerContent=self.readerContent.replace(importPlaceHolder,importString+importPlaceHolder)
			getMethodString=f'\tdef get{modelClassName}(self,forceRefresh=False):\r'
			getMethodString+=f'\t\trefreshJsonFilesLocallyIfOutdated(["completedforms"],forceRefresh)\r'
			getMethodString+=f'\t\tresult:List[{modelClassName}]=getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation("completedforms"),{modelClassName})\r'
			getMethodString+=f'\t\treturn result\r'
			self.readerContent+=getMethodString


import csv
from ..shared.helpers import ensureFolderExists,getPathRelativeToSkynamoDataFolder
from typing import List
def writeListOfObjectsToCsvWithObjectPropertiesAsColumnNames(listOfObjects:list, filename: str,columnsOrder:List[str]=[],delimiter: str = ',') -> None:
	##ensure filename is in output folder
	if 'output/' != filename[:7]:
		filename = 'output/' + filename
	##ensure filename ends with .csv
	if '.csv' != filename[-4:]:
		filename = filename + '.csv'
	headers=[]
	for col in columnsOrder:
		headers.append(col)
	for key in listOfObjects[0].__dict__.keys():
		if key not in headers:
			headers.append(key)
	ensureFolderExists(getPathRelativeToSkynamoDataFolder('output'))
	with open(getPathRelativeToSkynamoDataFolder(filename), 'w', newline='') as csvfile:
		writer = csv.DictWriter(csvfile, fieldnames=headers, delimiter=delimiter,quotechar='"', quoting=csv.QUOTE_ALL)
		writer.writeheader()
		for obj in listOfObjects:
			writer.writerow(obj.__dict__)


from .WriteOperationCls import WriteOperation
from .writeHelpers import getWriteOperationToUpdateObject,getBodyForWriteOperation,addWarehouseAndLabelToStockLevelUpdateIfPresent
from .execute import executeWrites
from typing import Union,Literal,Union
from ..models.Invoice import Invoice
from ..models.InvoiceItem import InvoiceItem
from datetime import datetime
from typing import List
from ..models.CustomFieldsToCreate import CustomFieldsToCreate
from ..models.Warehouse import Warehouse
from ..models.TaxRate import TaxRate
from ..models.PriceList import PriceList
from ..models.VisitFrequency import VisitFrequency


class WriterBase:
	def __init__(self):
		self.writeOperations:List[WriteOperation]=[]
	def apply(self):
		res= executeWrites(self.writeOperations)
		self.writeOperations=[]
		return res
	##unchanging update operations
	def addStockLevelUpdate(self,product_id:int,order_unit_id:int,level:int,warehouse_id:int=0,label:Union[None,str]=None):
		item={'product_id': product_id, 'order_unit_id': order_unit_id, 'level': level}
		addWarehouseAndLabelToStockLevelUpdateIfPresent(item,warehouse_id,label)
		self.writeOperations.append(WriteOperation("stocklevels", "post", item))

	def addStockLevelUpdateUsingProductCodeAndUnitName(self,product_code:str,order_unit_name:str,level:int,warehouse_id:int=0,label:Union[None,str]=None):
		item={'product_code': product_code, 'order_unit_name': order_unit_name, 'level': level}
		addWarehouseAndLabelToStockLevelUpdateIfPresent(item,warehouse_id,label)
		self.writeOperations.append(WriteOperation("stocklevels", "post", item))

	def addPriceUpdate(self,product_id:int,order_unit_id:int,price:float,price_list_id:int,tax_rate_id:Union[None,int]=None):
		item= {'product_id': product_id, 'order_unit_id': order_unit_id, 'price': price, 'price_list_id': price_list_id}
		if tax_rate_id!=None:
			item['tax_rate_id']=tax_rate_id
		self.writeOperations.append(WriteOperation("prices", "post",item))

	def addPriceUpdateUsingProductCodeAndUnitName(self,product_code:str,order_unit_name:str,price:float,price_list_id:int,tax_rate_id:Union[None,int]=None):
		item= {'product_code': product_code, 'order_unit_name': order_unit_name, 'price': price, 'price_list_id': price_list_id}
		if tax_rate_id!=None:
			item['tax_rate_id']=tax_rate_id
		self.writeOperations.append(WriteOperation("prices", "post",item))

	def addInvoiceUpdate(self,invoice:Invoice,fieldsToUpdate:List[str]):
		self.writeOperations.append(getWriteOperationToUpdateObject(invoice,fieldsToUpdate))
	##unchanging create operations
	def addInvoiceCreate(self,date:datetime,customer_code:str,items:List[InvoiceItem],reference='',status:Union[None,Literal['Draft','Authorized','Delivered','Outstanding','Paid','Deleted']]=None,due_date:Union[None,datetime]=None,taxIsIncludedInLineValues=True,outstanding_balance:Union[None,float]=None):
		body=getBodyForWriteOperation(locals())
		del body['taxIsIncludedInLineValues']
		if not(taxIsIncludedInLineValues):
			for item in body['items']:
				taxAmount=0
				if 'tax_amount' in item:
					taxAmount=item['tax_amount']
				item['value']=item['value']+taxAmount
		self.writeOperations.append(WriteOperation("invoices", "post", body))

	def addScheduledVisitCreate(self,assigned_user_name:str,customer_code:str,due_date:datetime,end_time:Union[datetime,None]=None,comment:Union[None,str]=None):
		body=getBodyForWriteOperation(locals())
		if end_time==None:
			body['all_day']=True
		self.writeOperations.append(WriteOperation("scheduledvisits", "post", body))

	def addCustomFieldCreations(self,customFieldsToCreate:CustomFieldsToCreate):
		self.writeOperations.append(WriteOperation("integrations", "post", {'action':'AddCustomFields','fields_to_add':customFieldsToCreate.fields_to_add},canBeCombinedWithOtherWritesInAList=False))
	## warehouses
	def addWarehouseCreate(self,name:str,order_email:str='',credit_request_email:str='',quote_email:str='',active:bool=True):
		body=getBodyForWriteOperation(locals())
		self.writeOperations.append(WriteOperation("warehouses", "post", body))

	def addWarehouseUpdate(self,warehouse:Warehouse,fieldsToUpdate:List[Literal['name','order_email','credit_request_email','quote_email','active']]):
		body={'id':warehouse.id,'name':warehouse.name}
		for field in fieldsToUpdate:
			body[field]=getattr(warehouse,field)
		self.writeOperations.append(WriteOperation("warehouses", "patch", body))
	## taxes
	def addTaxRateCreate(self,name:str,rate:float,active:bool=True):
		if rate<0 or rate>100:
			raise ValueError('rate must be between 0 and 100')
		body=getBodyForWriteOperation(locals())
		self.writeOperations.append(WriteOperation("taxrates", "post", body))
	def addTaxRateUpdate(self,taxrate:TaxRate,fieldsToUpdate:List[Literal['name','rate','active']]):
		body={'id':taxrate.id,'name':taxrate.name}
		for field in fieldsToUpdate:
			body[field]=getattr(taxrate,field)
		self.writeOperations.append(WriteOperation("taxrates", "patch", body))
	## pricelists
	def addPriceListCreate(self,name:str,prices_include_vat:bool,active:bool=True):
		body=getBodyForWriteOperation(locals())
		self.writeOperations.append(WriteOperation("pricelists", "post", body))
	def addPriceListUpdate(self,pricelist:PriceList,fieldsToUpdate:List[Literal['name','prices_include_vat','active']]):
		body={'id':pricelist.id,'name':pricelist.name}
		for field in fieldsToUpdate:
			body[field]=getattr(pricelist,field)
		self.writeOperations.append(WriteOperation("pricelists", "patch", body))
	## visitfrequencies
	def addVisitFrequencyCreate(self,customer_id:int,user_name:str,numberOfCyclesPerPeriod:int,numberOfVisitsRequiredPerCycle:int,period:Literal['Week','Month','Year']):
		body=getBodyForWriteOperation(locals())
		## replace numberOfCyclesPerPeriod with cycle and replace numberOfVisitsRequiredPerCycle with frequency
		body['cycle']=body['numberOfCyclesPerPeriod']
		body['frequency']=body['numberOfVisitsRequiredPerCycle']
		del body['numberOfCyclesPerPeriod']
		del body['numberOfVisitsRequiredPerCycle']
		self.writeOperations.append(WriteOperation("visitfrequencies", "post", body))
	
	def addVisitFrequencyUpdate(self,visitFrequency:VisitFrequency,fieldsToUpdate:List[Literal['customer_id','user_name','numberOfCyclesPerPeriod','numberOfVisitsRequiredPerCycle','period']]):
		body={'id':visitFrequency.id}
		for field in fieldsToUpdate:
			body[field]=getattr(visitFrequency,field)
		## replace numberOfCyclesPerPeriod with cycle and replace numberOfVisitsRequiredPerCycle with frequency
		if 'numberOfCyclesPerPeriod' in body:
			body['cycle']=body['numberOfCyclesPerPeriod']
			del body['numberOfCyclesPerPeriod']
		if 'numberOfVisitsRequiredPerCycle' in body:
			body['frequency']=body['numberOfVisitsRequiredPerCycle']
			del body['numberOfVisitsRequiredPerCycle']
		self.writeOperations.append(WriteOperation("visitfrequencies", "patch", body))




from typing import Union,List
from datetime import datetime
from ..shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from .LineItem import LineItem
from skynamo.reader.readerHelpers import getCustomFieldWithName

class Transaction:
	def getCustomFieldWithName(self,name:str):
		return getCustomFieldWithName(self,name)
	def __init__(self,json:dict):
		self.id:int=json['id']
		self.date:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['date'])
		self.customer_id:int=json['customer_id']
		self.customer_code:str=json['customer_code']
		self.customer_name:str=json['customer_name']
		self.reference:Union[str,None]=json['reference']
		self.interaction_id:int=json['interaction_id']
		self.discount_percentage:float=json['discount']
		self.discount_amount:float=json['discount_amount']
		self.total_amount:float=json['total_amount']
		self.prices_include_vat:Union[bool,None]=json['prices_include_vat']
		self.warehouse_id:Union[int,None]=json['warehouse_id']
		self.warehouse_name:Union[str,None]=json['warehouse_name']
		self.email_recipients:Union[str,None]=json['email_recipients']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])
		self.items:List[LineItem]=[]
		for item in json['items']:
			inputDict={'multiplier':1,'product_code':item['product_code'],'quantity':item['quantity'],'unit_name':item['order_unit_name'],'unit_price':item['unit_price'],'list_price':item['list_price'],'tax_rate_value':item['tax_rate_value'],'tax_rate_id':item['tax_rate_id']}
			lineItem=LineItem(**inputDict)
			lineItem.multiplier=None
			lineItem.product_name=item['product_name']
			self.items.append(lineItem)

		## the following properties are populated from completed forms after the object is created:
		self.user_id:int=0
		self.user_name:str=''




from datetime import datetime
from skynamo.shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from skynamo.reader.readerHelpers import getCustomFieldWithName

class CustomFormBase:
	def getCustomFieldWithName(self,name:str):
		return getCustomFieldWithName(self,name)
	def __init__(self,json:dict):
		self.id:int=json['id']
		self.date:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['date'])
		self.customer_id:int=json['customer_id']
		self.customer_code:str=json['customer_code']
		self.customer_name:str=json['customer_name']
		self.user_id:int=json['user_id']
		self.user_name:str=json['user_name']
		self.interaction_id:int=json['interaction_id']
		self.form_id:int=json['form_id']
		self.form_name:str=json['form_name']


from skynamo.models.Invoice import Invoice
from skynamo.models.User import User
from skynamo.models.Warehouse import Warehouse
from skynamo.models.TaxRate import TaxRate
from skynamo.models.Price import Price
from skynamo.models.StockLevel import StockLevel
from skynamo.models.PriceList import PriceList
from skynamo.models.VisitFrequency import VisitFrequency
from skynamo.reader.sync import refreshJsonFilesLocallyIfOutdated,getSynchedDataTypeFileLocation
from skynamo.reader.jsonToObjects import getListOfObjectsFromJsonFile,populateCustomPropsFromFormResults,populateUserIdAndNameFromInteractionAndReturnFormIds
from typing import List

class ReaderBase:
	def __init__(self):
		pass
	def getInvoices(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['invoices'],forceRefresh)
		invoices:List[Invoice]=getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('invoices'),Invoice)
		return invoices

	def getStockLevels(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['stocklevels'],forceRefresh)
		stockLevels:List[StockLevel]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('stocklevels'),StockLevel)
		return stockLevels

	def getUsers(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['users'],forceRefresh)
		users:List[User]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('users'),User)
		return users

	def getWarehouses(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['warehouses'],forceRefresh)
		warehouses:List[Warehouse]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('warehouses'),Warehouse)
		return warehouses

	def getPrices(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['prices'],forceRefresh)
		prices:List[Price]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('prices'),Price)
		return prices
	
	def getTaxRates(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['taxrates'],forceRefresh)
		taxRates:List[TaxRate]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('taxrates'),TaxRate)
		return taxRates
	
	def getPriceLists(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['pricelists'],forceRefresh)
		priceLists:List[PriceList]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('pricelists'),PriceList)
		return priceLists
	
	def getVisitFrequencies(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['visitfrequencies'],forceRefresh)
		visitFrequencies:List[VisitFrequency]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('visitfrequencies'),VisitFrequency)
		return visitFrequencies


from .CreditRequest import CreditRequest
from .Order import Order
from .Quote import Quote
from .Product import Product
from .Customer import Customer

from skynamo.reader.jsonToObjects import getListOfObjectsFromJsonFile,populateCustomPropsFromFormResults,populateUserIdAndNameFromInteractionAndReturnFormIds
from skynamo.reader.sync import refreshJsonFilesLocallyIfOutdated,getSynchedDataTypeFileLocation
import json
from typing import List
from skynamo.reader.ReaderBase import ReaderBase
##|customImports|##

def _getTransactions(transactionClass,forceRefresh=False):
	refreshJsonFilesLocallyIfOutdated([f'{transactionClass.__name__.lower()}s','completedforms','interactions'])#type:ignore
	interactionsJson={}
	with open(getSynchedDataTypeFileLocation('interactions'), "r") as read_file:
		interactionsJson=json.load(read_file)
	completedForms={}
	with open(getSynchedDataTypeFileLocation('completedforms'), "r") as read_file:
		completedForms=json.load(read_file)
	transactions=getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation(f'{transactionClass.__name__.lower()}s'),transactionClass)
	populatedTransactions=[]
	for i,transaction in enumerate(transactions):
		formIds=populateUserIdAndNameFromInteractionAndReturnFormIds(transaction,interactionsJson)
		try:
			populateCustomPropsFromFormResults(transaction,formIds,completedForms)
			populatedTransactions.append(transaction)
		except Exception as e:
			print(f'Warning: Error populating custom props for {transactionClass.__name__} {i}: {e}. Leaving out of transactions list since it might be because since the last sync new orders have come through but not yet their associated form results.')
	return populatedTransactions

class Reader(ReaderBase):
	def __init__(self):
		pass
	def getOrders(self,forceRefresh=False):
		orders:List[Order]=_getTransactions(Order,forceRefresh)
		return orders

	def getCreditRequests(self,forceRefresh=False):
		creditRequests:List[CreditRequest]=_getTransactions(CreditRequest,forceRefresh)
		return creditRequests

	def getQuotes(self,forceRefresh=False):
		quotes:List[Quote]=_getTransactions(Quote,forceRefresh)
		return quotes

	def getProducts(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['products'],forceRefresh)
		products:List[Product]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('products'),Product)
		return products

	def getCustomers(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['customers'],forceRefresh)
		customers:List[Customer]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('customers'),Customer)
		return customers

	




import os
import requests
from typing import Literal,Any,Union,Dict
from .helpers import updateEnvironmentVariablesFromJsonConfig

def getApiBase():
	region=os.environ.get('SKYNAMO_REGION')
	if region==None:
		updateEnvironmentVariablesFromJsonConfig()
		region=os.environ.get('SKYNAMO_REGION')
	return f'https://api.{region}.skynamo.me/v1/'

def getHeaders():
	instanceName=os.environ.get('SKYNAMO_INSTANCE_NAME')
	apiKey=os.environ.get('SKYNAMO_API_KEY')
	if instanceName==None or apiKey==None:
		updateEnvironmentVariablesFromJsonConfig()
		instanceName=os.environ.get('SKYNAMO_INSTANCE_NAME')
		apiKey=os.environ.get('SKYNAMO_API_KEY')
	return {'x-api-client':instanceName,'x-api-key':apiKey,'accept':'application/json'}

def makeRequest(method:Literal['get','post','patch','put'],dataType:str,dataOrParams:Union[list,Dict[str,Any]]={}):
	print(method)
	print(dataType)
	print(dataOrParams)
	if method=='get':
		return requests.get(getApiBase()+dataType,headers=getHeaders(),params=dataOrParams).json()
	elif method=='post':
		return requests.post(getApiBase()+dataType,headers=getHeaders(),json=dataOrParams).json()
	elif method=='patch':
		return requests.patch(getApiBase()+dataType,headers=getHeaders(),json=dataOrParams).json()
	elif method=='put':
		return requests.put(getApiBase()+dataType,headers=getHeaders(),json=dataOrParams).json()
	elif method=='delete':
		return requests.delete(getApiBase()+dataType,headers=getHeaders(),json=dataOrParams).json()
	else:
		raise Exception(f'Invalid method: {method}')



from typing import Union
class Location:
	def __init__(self,latitude=0.0,longitude=0.0,accuracy=0.0,is_approximate=False):
		self.latitude=latitude
		self.longitude=longitude
		self.accuracy=accuracy
		self.is_approximate=is_approximate
	def getJsonReadyValue(self):
		if self.latitude==0.0 and self.longitude==0.0:
			return None
		return {'latitude':self.latitude,'longitude':self.longitude,'accuracy':self.accuracy,'is_approximate':self.is_approximate}



def getCustomFieldWithName(skynamoObject,name:str):
	nameInProps=name.replace(' ','_')
	## if prop starts with c and ends with f'_{nameInProps}
	for prop in skynamoObject.__dict__:
		if prop[0] in ['c','f'] and f'_{nameInProps}'==prop[-len(nameInProps)-1:]:
			return skynamoObject.__dict__[prop]
	return None


class Address:
	def __init__(self,street:str='',city:str='',state:str='',zip:str=''):
		self.street=street
		self.city=city
		self.state=state
		self.zip=zip

	def populateFromJsonValue(self,jsonValue):
		##split on newline
		lines=jsonValue.split('\n')
		for i in range(4):
			if i<len(lines):
				if i==0:
					self.street=lines[i]
				elif i==1:
					self.city=lines[i]
				elif i==2:
					self.state=lines[i]
				elif i==3:
					self.zip=lines[i]

	def getJsonReadyValue(self):
		return f'{self.street}\n{self.city}\n{self.state}\n{self.zip}'


from typing import List,Dict,Any
from skynamo.main.Reader import Reader
from skynamo.main.Writer import Writer
from skynamo.models.TaxRate import TaxRate
from skynamo.models.Warehouse import Warehouse

def syncTaxRatesWithSkynamoAndReturnNameLookup(taxRates:List[Dict[str,Any]],idField:str,nameField:str,rateField:str,isActiveField:str='deleted_at',isActiveValue:Any='',multiplyRateBy=1):
	erpTaxRateIdToSkynamoTaxRates=getDictOfErpTaxRateIdToSkynamoTaxRate()
	writer=Writer()
	nameLookup={}
	for erpTaxRate in taxRates:
		erpTaxRateId=erpTaxRate[idField]
		name=erpTaxRate[nameField]+' ('+erpTaxRateId+')'
		nameLookup[erpTaxRateId]=name
		newRate=float(erpTaxRate[rateField])*multiplyRateBy
		isActiveInErp=erpTaxRate[isActiveField]==isActiveValue
		if erpTaxRateId not in erpTaxRateIdToSkynamoTaxRates:# if tax rate is not in skynamo and is active in erp
			if isActiveInErp:
				writer.addTaxRateCreate(name,newRate)
		else:# if tax rate is in skynamo
			skynamoTaxRate=erpTaxRateIdToSkynamoTaxRates[erpTaxRateId]
			if name!=skynamoTaxRate.name or newRate!=skynamoTaxRate.rate or isActiveInErp!=skynamoTaxRate.active:
				skynamoTaxRate.name=name
				skynamoTaxRate.rate=newRate
				skynamoTaxRate.active=isActiveInErp
				writer.addTaxRateUpdate(skynamoTaxRate,['name','rate','active'])
	errors=writer.apply()
	if errors!=[]:
		raise Exception('Error syncing tax rates with Skynamo: '+str(errors))
	return nameLookup

def syncWarehousesWithSkynamoAndReturnNameLookup(erpWarehouses:List[Dict[str,Any]],idField:str,nameField:str,isActiveField:str='deleted_at',isActiveValue:Any=''):
	erpWarehouseIdToSkynamoWarehouse=getDictOfErpWarehouseIdToSkynamoWarehouse()
	writer=Writer()
	nameLookup={}
	for erpWarehouse in erpWarehouses:
		erpWarehouseId=erpWarehouse[idField]
		name=erpWarehouse[nameField]+' ('+erpWarehouseId+')'
		nameLookup[erpWarehouseId]=name
		isActiveInErp=erpWarehouse[isActiveField]==isActiveValue
		if erpWarehouseId not in erpWarehouseIdToSkynamoWarehouse:# if warehouse is not in skynamo and is active in erp
			if isActiveInErp:
				writer.addWarehouseCreate(name)
		else:
			skynamoWarehouse=erpWarehouseIdToSkynamoWarehouse[erpWarehouseId]
			if name!=skynamoWarehouse.name or isActiveInErp!=skynamoWarehouse.active:
				skynamoWarehouse.name=name
				skynamoWarehouse.active=isActiveInErp
				writer.addWarehouseUpdate(skynamoWarehouse,['name','active'])
	errors=writer.apply()
	if errors!=[]:
		raise Exception('Error syncing warehouses with Skynamo: '+str(errors))
	return nameLookup


def getDictOfErpTaxRateIdToSkynamoTaxRate():
	reader=Reader()
	result:Dict[str,TaxRate]={}
	existingTaxRates=reader.getTaxRates(forceRefresh=True)
	for taxRate in existingTaxRates:
		addTaxOrWarehouseObjectToResult(result,taxRate)
	return result

def getDictOfErpWarehouseIdToSkynamoWarehouse():
	reader=Reader()
	result:Dict[str,Warehouse]={}
	existingWarehouses=reader.getWarehouses(forceRefresh=True)
	for warehouse in existingWarehouses:
		addTaxOrWarehouseObjectToResult(result,warehouse)
	return result

def addTaxOrWarehouseObjectToResult(result:Dict[str,Any],object):
	firstBracketPos=object.name.find('(')
	lastBracketPos=object.name.find(')')
	if firstBracketPos!=-1 and lastBracketPos!=-1 and lastBracketPos>firstBracketPos:
		erpTaxRateId=object.name[firstBracketPos+1:lastBracketPos]
		result[erpTaxRateId]=object


from typing import Union,Literal,List,Dict
from datetime import datetime
from skynamo.shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from skynamo.models.Address import Address
from skynamo.models.OrderUnit import OrderUnit
from skynamo.reader.readerHelpers import getCustomFieldWithName

class Product:
	def getCustomFieldWithName(self,name:str):
		return getCustomFieldWithName(self,name)
	def __init__(self,json:dict={}):
		self.id:str=json['id']
		self.row_version:int=json['row_version']
		self.code:str=json['code']
		self.name:str=json['name']
		self.active:bool=json['active']
		self.order_units:List[OrderUnit]=[]
		for order_unit in json['order_units']:
			id=None
			if 'id' in order_unit:
				id=order_unit['id']
			self.order_units.append(OrderUnit(order_unit['name'],order_unit['multiplier'],id=id))
			if 'active' in order_unit:
				self.order_units[-1].active=order_unit['active']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])



from typing import Literal,List,Union
class WriteError:
	def __init__(self,dataType:str,httpMethod:str,body:Union[list,dict,str],error:List[str]):
		self.dataType=dataType
		self.body=body
		self.httpMethod=httpMethod
		self.error=error


from ..shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr

class TaxRate:
    def __init__(self,json:dict):
        self.id=int(json['id'])
        self.name=str(json['name'])
        self.rate=float(json['rate'])
        self.active=bool(json['active'])
        self.last_modified_time=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])


from typing import Union

class LineItem:
	def __init__(self,product_code:str,quantity:float,unit_name:str,multiplier:int,unit_price:float,list_price:Union[None,float]=None,tax_rate_value:Union[float,None]=None,tax_rate_id:Union[int,None]=None):
		self.product_code=product_code
		self.quantity=quantity
		self.unit_price:float=unit_price
		self.list_price=list_price
		self.unit_name=unit_name
		self.multiplier:Union[None,float]=multiplier
		self.tax_rate_value=tax_rate_value
		self.tax_rate_id=tax_rate_id
		self.product_name:Union[str,None]=None

	def getJsonReadyValue(self):
		return self.__dict__


from typing import Literal, Union

class WriteOperation:
	def __init__(self,dataType:str,httpMethod:Literal['post','patch','put','delete'],itemOrId:Union[dict,str],canBeCombinedWithOtherWritesInAList=True):
		self.dataType=dataType
		self.httpMethod=httpMethod
		self.itemOrId=itemOrId
		self.canBeCombinedWithOtherWritesInAList=canBeCombinedWithOtherWritesInAList


class PriceList:
    def __init__(self,json:dict):
        self.id=json['id']
        self.name=json['name']
        self.last_modified_time=json['last_modified_time']
        self.active=json['active']
        self.prices_include_vat=json['prices_include_vat']


from ..shared.helpers import updateEnvironmentVariablesFromJsonConfig,getPathRelativeToSkynamoDataFolder
from typing import List

def sendEmailUsingGmailCredentialsWithFilesAttached(subject: str, body: str, recipients: List[str], files: list) -> None:
	import smtplib
	from email.mime.multipart import MIMEMultipart
	from email.mime.text import MIMEText
	from email.mime.base import MIMEBase
	from email import encoders
	from email.mime.application import MIMEApplication

	## get sender and sender password from environment variables
	import os
	if 'SKYNAMO_GMAIL_SENDER' not in os.environ:
		updateEnvironmentVariablesFromJsonConfig()
	sender=os.environ.get('SKYNAMO_GMAIL_SENDER')
	senderPassword=os.environ.get('SKYNAMO_GMAIL_PASSWORD')
	if sender is None or senderPassword is None:
		raise Exception('SKYNAMO_GMAIL_SENDER and SKYNAMO_GMAIL_PASSWORD environment variables must be set to send emails using Gmail credentials.')
	msg = MIMEMultipart()
	msg['From'] = sender
	msg['To'] = ','.join(recipients)
	msg['Subject'] = subject
	msg.attach(MIMEText(body, 'plain'))

	for file in files:
		if 'output/' != file[:7]:
			file = 'output/'+file
		with open(getPathRelativeToSkynamoDataFolder(file), "rb") as fil:
			part = MIMEApplication(
				fil.read(),
				Name=file.split('/')[-1]
			)
		# After the file is closed
		part['Content-Disposition'] = 'attachment; filename="%s"' % file.split('/')[-1]
		msg.attach(part)

	server = smtplib.SMTP('smtp.gmail.com', 587)
	server.starttls()
	server.login(sender, senderPassword)
	text = msg.as_string()
	server.sendmail(sender, recipients, text)
	server.quit()


from typing import Union,Literal,List,Dict
from datetime import datetime
from skynamo.models.Address import Address
from skynamo.shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from skynamo.models.Transaction import Transaction

class CreditRequest(Transaction):
	def __init__(self,json:dict):
		super().__init__(json)
